/*=====================================================================
profiler.cpp
------------
File created by ClassTemplate on Thu Feb 24 19:00:30 2005

Copyright (C) Nicholas Chapman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

http://www.gnu.org/copyleft/gpl.html..
=====================================================================*/
#include "profiler.h"


#include "../utils/stringutils.h"
#include <process.h>
#include <iostream>
#include <assert.h>
#include <winnt.h>
#include <dbghelp.h>

// DE: 20090325: Profiler no longer owns callstack and flatcounts since it is shared between multipler profilers

Profiler::Profiler(HANDLE target_process_, HANDLE target_thread_, std::map<CallStack, int>& callstacks_, std::map<PROFILER_ADDR, int>& flatcounts_)
:	target_process(target_process_),
	target_thread(target_thread_),
	callstacks(callstacks_),
	flatcounts(flatcounts_)
{
	
}

// DE: 20090325: Need copy constructor since it is put in a std::vector

Profiler::Profiler(const Profiler& iOther)
:	target_process(iOther.target_process),
	target_thread(iOther.target_thread),
	callstacks(iOther.callstacks),
	flatcounts(iOther.flatcounts)
{
}

// DE: 20090325: Need copy assignement since it is put in a std::vector

Profiler& Profiler::operator=(const Profiler& iOther)
{
	target_process = iOther.target_process;
	target_thread = iOther.target_thread;
	callstacks = iOther.callstacks;
	flatcounts = iOther.flatcounts;

	return *this;
}

Profiler::~Profiler()
{
	
}


bool Profiler::sampleTarget()
{
	// DE: 20090325: Moved declaration of stack variables to reduce size of code inside Suspend/Resume thread

	CallStack stack;
	stack.depth = 0;

	STACKFRAME64 frame;
	memset(&frame, 0, sizeof(frame));

	CONTEXT threadcontext;
	threadcontext.ContextFlags = CONTEXT_i386 | CONTEXT_CONTROL;

	// An open question is whther or not this routine can be called
	// reentrantly by the multi-media timer support.

	// Can fail occasionally, for example if you have a debugger attached to the process.
	HRESULT result = SuspendThread(target_thread);
	if(result == 0xffffffff)
		return false;

	int prev_priority = GetThreadPriority(target_thread);
	
	SetThreadPriority(target_thread, THREAD_PRIORITY_TIME_CRITICAL);
	result = GetThreadContext(target_thread, &threadcontext);
	SetThreadPriority(target_thread, prev_priority);

	if(!result){
		// DE: 20090325: If GetThreadContext fails we must be sure to resume thread again
		ResumeThread(target_thread);
		throw ProfilerExcep("GetThreadContext failed.");
	}

	stack.addr[stack.depth++] = threadcontext.Eip;
	frame.AddrStack.Offset = threadcontext.Esp;
	frame.AddrPC.Offset = threadcontext.Eip;
	frame.AddrFrame.Offset = threadcontext.Ebp;
	frame.AddrStack.Mode = frame.AddrPC.Mode = frame.AddrFrame.Mode = AddrModeFlat;

	while(true)
	{
		BOOL result = StackWalk64(
			IMAGE_FILE_MACHINE_I386,
			target_process,
			target_thread,
			&frame,
			&threadcontext,
			NULL,
			&SymFunctionTableAccess64,
			&SymGetModuleBase64,
			NULL
		);

		if (!result || stack.depth >= MAX_CALLSTACK_LEVELS || frame.AddrReturn.Offset == 0)
			break;

		stack.addr[stack.depth++] = (PROFILER_ADDR)frame.AddrReturn.Offset;
	}

	//std::cout << "addr: " << addr << std::endl;

	result = ResumeThread(target_thread);
	if(!result)
		throw ProfilerExcep("ResumeThread failed.");

	//NOTE: this has to go after ResumeThread.  Otherwise mem allocation needed by std::map
	//may hit a lock held by the suspended thread.

	flatcounts[stack.addr[0]]++;
	callstacks[stack]++;
	return true;
}

// returns true if the target thread has finished
bool Profiler::targetExited() const
{
	DWORD code = WaitForSingleObject(target_thread, 0);
	return (code == WAIT_OBJECT_0);
}


//void Profiler::saveIPs(std::ostream& stream)
//{
//	for(std::map<Sample, int>::const_iterator i = counts.begin(); 
//		i != counts.end(); ++i)
//	{
//		const Sample &sample = i->first;
//		int count = i->second;
//		stream << ::toHexString(sample.addr) << " " << count << "\n";
//	}
//
//	stream.flush();
//}
//
